// cygtest.cpp : Defines the entry point for the console application.
//


#include <SDKDDKVer.h>
#include <stdio.h>
#include <windows.h>
#include <string>
using namespace std;

LPSTR __stdcall UnicodeToMByteHelper(LPSTR lpa, int nBytes, LPCWSTR lpw, int nChars, int codepage);

static UINT cyg_codepage_string_to_CP(const string &cp)
{
  const string UTF8 = "UTF-8";
  const string utf8 = "utf-8";
  const string ANSI = "ANSI";
  const string ansi = "ansi";
  const string ISO88591 = "ISO-8859-1";
  const string iso88591 = "iso-8859-1";
  const string OEM = "OEM";
  const string oem = "oem";
  const string WINDOWS = "WINDOWS";
  const string windows = "windows";
  const string CODEPAGE = "CP";
  const string codepage = "cp";
  UINT shell_cp{ 0 };

  if (NULL == cp.c_str() || cp.length() == 0)
    return 0;

  if ((cp.compare(utf8) == 0) || (cp.compare(UTF8) == 0))
    shell_cp = 65001;
  else if ((cp.compare(ansi) == 0) || (cp.compare(ANSI) == 0)
    || (cp.compare(ISO88591) == 0) || (cp.compare(iso88591) == 0))
    shell_cp = 1252;
  // oem is also standard cygwin nomenclature
  else if ((cp.compare(oem) == 0) || (cp.compare(OEM) == 0))
    shell_cp = 437;
  // cpXXX, windows-XXX and windows_XXX are all recognized by
  // the Ab Initio extensions to cygwin.  Not sure if they are
  // known to standard cygwin, but I don't think they are.
  else if ((cp.compare(0, 2, codepage) == 0) ||
    (cp.compare(0, 2, CODEPAGE) == 0) ||
    (cp.compare(0, 7, windows) == 0) ||
    (cp.compare(0, 7, WINDOWS) == 0)) {
    // If the prefix is "CP" or "cp" then get the number after that
    // else it's "WINDOWS{-,_}" or "WINDOWS{-,_}"
    int offset = ((cp.compare(0, 2, codepage) == 0) || (cp.compare(0, 2, CODEPAGE) == 0)) ? 2 : 8;
    shell_cp = atoi(cp.substr(offset).c_str());
  }
  return shell_cp;
}

static UINT get_cygwin_codepage()
{
  string default_cyg_charset = "C.UTF-8";        // Cygwin default character set
  string cyg_locale;
  UINT shell_cp{ 0 };
  UINT default_cp{ 65001 };
  char *envptr = ::getenv("LANG");

  if (NULL == envptr)
    envptr = ::getenv("LC_ALL");

  cyg_locale = (NULL == envptr ? default_cyg_charset : envptr);
  // The 'value' field of the environment string "var_name=value"
  // will be of the form: <language ID>.<codepage ID>
  // We want the substring after the '.'  
  int dotPos = cyg_locale.find_first_of('.');
  if (dotPos >= 0) {
    // The character set string, if specified, starts AFTER  the '.'.
    // If NOT specified, return the input default.
    string page = cyg_locale.substr(++dotPos);
    if (0 <= (shell_cp = cyg_codepage_string_to_CP(page))) {
      return shell_cp;
    }  // end SHELL_CP
  }    // end EQPOS
  return default_cp;
}


LPSTR __stdcall UnicodeToMByteHelper(LPSTR lpa, int nBytes, LPCWSTR lpw, int nChars, int codepage)
{
  static int printInfo = 0;
  int nOut = 0;

  if (NULL == lpa) {
    printf("NULL input string\n");
    return NULL;
  }

  if (printInfo) {
    printf("Transcoding using Cygwin codepage: %d\nInput widechar string:\n", codepage);
    for (int i = 0; i < nChars; i++)
      printf("\tlpw[%d] = %C - %02X\n", i, lpw[i], lpw[i]);
  }
  ++printInfo;

  if (nChars > 0) {
    if (0 == (nOut = WideCharToMultiByte(codepage, 0, lpw, nChars, lpa, nBytes, NULL, NULL))) {
      DWORD dwErr = GetLastError();
      printf("WideCharToMultiByte(%d, %S) failed, error %d\n", codepage, lpw, dwErr);
      return NULL;
    }
  }
  lpa[nOut] = '\0';
  return lpa;
}

int wmain(int argc, wchar_t** wargv)
{
  try {
    char *pNull = "NULL";
    char** argv = new char*[(argc)+1];
    int _argi;
    int codepage = get_cygwin_codepage();
    for (_argi = 0; _argi < (argc); _argi++) {
      if (wargv[_argi]) {
        LPWSTR utf_lpw  = wargv[_argi];
        int utf_len     = lstrlenW(utf_lpw);
        int utf_convert = utf_len * 3 + 1;
        LPSTR utf_lpa   = (LPSTR)_alloca(utf_convert);
        argv[_argi]     = UnicodeToMByteHelper(utf_lpa, utf_convert, utf_lpw, utf_len, codepage);
      }
      else {
        argv[_argi] = pNull;
      }
    }
    argv[(argc)] = NULL;

    // Now print the transcoded string.

    for (int i = 1; i < argc; i++)
      printf("%s: %s\n", __FUNCTION__, argv[i]);

    return 0;
  }
  catch (...) {
    printf("Caught unhandled exception\n");
  }
}
